/* $Id: pbm_toolkit.cpp,v 1.3 2004/05/09 01:36:19 chris Exp $
 *
 * shapsim - Amorphous shape mapping computer simulator and analyzer.
 * Copyright (C) 2004 Chris Frost <frost@cs.virginia.edu>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "pbm_toolkit.h"
#include <string>
#include <vector>
#include <fstream>
#include <iostream>
using namespace std;

pbm::pbm(const string &fn)
	: filename(fn)
{}


#define OPEN_PBM(filename) \
	ifstream smf(filename.c_str()); \
	if(!smf) \
	{ \
		cerr << "ERROR: Unable to open shape description file \"" \
			  << filename << "\", exiting." << endl; \
		exit(1); \
	} \

#define PBM_MAGIC_NUM "P1"

string pbm::get_magic_num()
{
	OPEN_PBM(filename)

	string magic_num;
	smf >> magic_num;

	smf.close();
	return magic_num;
}

bool pbm::correct_magic_num()
{
	return get_magic_num() == PBM_MAGIC_NUM;
}

unsigned int pbm::get_width()
{
	OPEN_PBM(filename)

	string magic_num;
   smf >> magic_num;
	assert(magic_num == PBM_MAGIC_NUM);


   unsigned int width;
	smf >> width;

	smf.close();
	return width;
}

unsigned int pbm::get_height()
{
	OPEN_PBM(filename)

	string magic_num;
   smf >> magic_num;
	assert(magic_num == PBM_MAGIC_NUM);

   unsigned int width;
	smf >> width;

		
   unsigned int height;
	smf >> height;

	smf.close();
	return height;
}

vector< vector<bool> > pbm::get_image()
{
	OPEN_PBM(filename)

	string magic_num;
   smf >> magic_num;
	assert(magic_num == PBM_MAGIC_NUM);

   unsigned int width;
	smf >> width;

   unsigned int height;
	smf >> height;


	bool pixel;
	char pixel_as_char;
	vector< vector<bool> > image_y_x;
	for(unsigned int y=0; y<height; ++y)
	{
		vector<bool> image_row;
		for(unsigned int x=0; x<width; ++x)
		{
			// skip over whitespace
			smf.get(pixel_as_char);
			while(pixel_as_char != '0' && pixel_as_char != '1')
				smf.get(pixel_as_char);

			if('0' == pixel_as_char)
				pixel = false;
			else
				pixel = true;

			image_row.push_back(pixel);
		}
		image_y_x.push_back(image_row);
	}

	smf.close();
	return image_y_x;
}

#include <stdio.h>
#include <stdlib.h>
#include <sys/wait.h>
#include "utils.h"
vector< vector<bool> > pbm::get_image(unsigned int width, unsigned int height)
{
	bool fatal_error = false;

	const unsigned int
		orig_width = get_width(),
		orig_height = get_height();

	if(orig_width == width && orig_height == height)
		return get_image();

	
	// WARNING: tmpnam()'s documentation says never to use it.
	// We do use it to make things easy here, and are safe as long as
	// tmpname() is not used elsewhere in this program, but it would be
	// good to use tmpfile() instead.
	const string scaled_image_filename(tmpnam(NULL));
	const string scale_cmd
		= "pnmscale -xysize " + uitos(width) + " " + uitos(height)
		+ " " + filename
		+ " | pgmtopbm | pnmtoplainpnm > " + scaled_image_filename;

	int scale_ret = system(scale_cmd.c_str());
	if(scale_ret != -1 && WEXITSTATUS(scale_ret) != 0)
	{
		cerr << "ERROR: Error scaling image using pnmscale, exiting." << endl;
		fatal_error = true; // continue so that we cleanup our tmpfile
	}

	pbm scaled_pbm(scaled_image_filename);
	vector< vector<bool> > scaled_image(scaled_pbm.get_image());

	int rm_ret = remove(scaled_image_filename.c_str());
	if(rm_ret != 0)
	{
		cerr << "ERROR: Error removing temp image for scaling, exiting." << endl;
		fatal_error = true;
	}

	if(fatal_error) exit(1);

	// FIXME: why does our assert here fail?
	// Theory: pnmscale doesn't work quite like I think, maybe it fits the
	// smaller axis?
	if(height != scaled_image.size())
		cerr << "scaled_image height: " << scaled_image.size() << endl;
	if(width  != scaled_image[0].size())
		cerr << "scaled_image width: " << scaled_image[0].size() << endl;
	// Removed asserts because someone could ask for a larger version of the img
	//	assert(height <= scaled_image.size());
	//	assert(width  <= scaled_image[0].size());
	
	return scaled_image;
}
