/* $Id: sca.cpp,v 1.3 2004/05/09 01:36:20 chris Exp $
 *
 * shapsim - Amorphous shape mapping computer simulator and analyzer.
 * Copyright (C) 2004 Chris Frost <frost@cs.virginia.edu>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "shape_cell_analyzer.h"
#include "shape_cell_querier.h"
#include "location.h"
#include "utils.h"
#include <string>
#include <iostream>
using namespace std;

// Usage:
// sca [output_file_basename]

int main(int argc, char **argv)
{
	const bool verbose_output = true;
	const int required_num_args = 3;

	if(required_num_args > argc)
	{
		cerr << "Usage: sca <query_and_output_filename_prefix>"
			  << " <shape_filename>"
			  << " [percent_cells_queried in [0,1]]"
			  << endl;
		exit(1);
	}

	string filename_prefix = argv[1];
	string shape_filename  = argv[2];
	float percent_cells_queried = 1;
	if(1+required_num_args == argc)
		percent_cells_queried = atof(argv[required_num_args]);

	vector<shape_data> shaped;
	location           space_size;
	location::coord_t  neighbor_max_dist;
	string             locn_nghbr_filename, shape_description_filename;
	unsigned int       prng_seed;

	//
	//
	// Load queried data
	if(verbose_output)
		cout << "Loading queried data:" << endl;
	assert(load_queried_data(filename_prefix+"-queried_data",
									 shaped, space_size, neighbor_max_dist,
									 locn_nghbr_filename, shape_description_filename,
									 prng_seed,
									 verbose_output));

	//
	// Invalidate the given percentage of cells, equivalent to only
	// sampling one minus this percent.
	const unsigned int num_cells_queried
		= decrease_cell_sample_size(percent_cells_queried, shaped);

	//
	//
	// Generate cell group size distribution
	if(verbose_output)
		cout << "Generating cell group size distribution statistics:" << endl;
	unsigned int member__largest_group;
	generate_connection_stats(shaped, filename_prefix+"-connections_stats",
									  member__largest_group, verbose_output);
	if(UINT_MAX == member__largest_group)
	{
		cout << "No cells queried, exiting." << endl;
		exit(0);
	}

	//
	//
	// Generate graph of node connections
	if(verbose_output)
		cout << "Generating connection graph:" << endl;
	generate_connection_graph(shaped, filename_prefix+"-connections.neato",
									  verbose_output);

	//
	//
	// Generate shape covering data
	vector<cell_cov> sim_shape_cov;
	if(verbose_output)
		cout << "Generating space coloring from simulator data:" << endl;
	generate_sim_shape_cov(shaped, sim_shape_cov, verbose_output);
	assert(output_shape_cov_coloring(sim_shape_cov, space_size,
												filename_prefix+"-shape_coloring-sim.ppm",
												verbose_output));

	vector<cell_cov> cell_shape_cov;
	location         cell_shape_cov_space_size;
	if(verbose_output)
		cout << "Generating space coloring from cell data:" << endl;
	generate_cell_shape_cov(shaped, member__largest_group,
									cell_shape_cov, cell_shape_cov_space_size,
									verbose_output);
	assert(output_shape_cov_coloring(cell_shape_cov, cell_shape_cov_space_size,
												filename_prefix+"-shape_coloring-cell.ppm",
												verbose_output));

	assert(calculate_shapecov_stats(cell_shape_cov, cell_shape_cov_space_size,
										  shape_filename,
										  filename_prefix+"-shape_coloring-cell-stats"));


	//
	//
	// Produce convex hulls
	if(verbose_output)
		cout << "Finding the convex hulls." << endl;

	//
	// Store points to compute convex hull with

	// Pull together only points needed for convex hull (to save runtime time)
	vector<cell_cov> convex_hull_cc_sim;
	remove_copy_if(sim_shape_cov.begin(), sim_shape_cov.end(),
						back_inserter(convex_hull_cc_sim),
						offshape_cell());
	vector<cell_cov> convex_hull_cc_cell;
	remove_copy_if(cell_shape_cov.begin(), cell_shape_cov.end(),
						back_inserter(convex_hull_cc_cell),
						not_boundary_cell());

	// Extract locations pass on to convex hull algo
	vector<location> convex_hull_frompts_sim, convex_hull_frompts_cell;
	transform(convex_hull_cc_sim.begin(), convex_hull_cc_sim.end(),
				 back_inserter(convex_hull_frompts_sim),
				 cell_cov2location());
	transform(convex_hull_cc_cell.begin(), convex_hull_cc_cell.end(),
				 back_inserter(convex_hull_frompts_cell),
				 cell_cov2location());

	//
	// Find the hulls
	vector<location> convex_hull_pts_sim, convex_hull_pts_cell;
	convex_hull(convex_hull_frompts_sim,  convex_hull_pts_sim);
	convex_hull(convex_hull_frompts_cell, convex_hull_pts_cell);

	output_locations(convex_hull_pts_sim, space_size,
						  filename_prefix+"-ch-sim.pgm", verbose_output);
	output_locations(convex_hull_pts_cell, cell_shape_cov_space_size,
						  filename_prefix+"-ch-cell.pgm", verbose_output);

	//
	// Calc difference in area of hulls
	const float sim_hull_area  = area(convex_hull_pts_sim);
	const float cell_hull_area = area(convex_hull_pts_cell);
	const float hull_area_diff = cell_hull_area - sim_hull_area;
	assert(output_string(ftos(hull_area_diff)
								+ " / " + ftos(sim_hull_area)
								+ " = " + ftos(hull_area_diff / sim_hull_area),
								filename_prefix+"-ch_diff"));

	//
	// Calc sampled point correctness
	if(verbose_output)
		cout << "Sampling points within polygonal description." << endl;

	const int num_samples = 100000;
	float correct_in,correct_out,correct, incorrect_in,incorrect_out,incorrect;
	calc_percent_inout(convex_hull_pts_sim, convex_hull_pts_cell,
							 space_size, num_samples,
							 correct_in,   correct_out,   correct,
							 incorrect_in, incorrect_out, incorrect,
							 verbose_output);
	assert(output_string(ftos(correct_in)
								+ " " + ftos(correct_out)
								+ " " + ftos(correct)
								+ " " + ftos(incorrect_in)
								+ " " + ftos(incorrect_out)
								+ " " + ftos(incorrect),
								filename_prefix+"-ch_samples"));

	return 0;
}
