/* $Id: scheduler.cpp,v 1.4 2004/05/11 12:42:49 chris Exp $
 *
 * shapsim - Amorphous shape mapping computer simulator and analyzer.
 * Copyright (C) 2004 Chris Frost <frost@cs.virginia.edu>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "scheduler.h"
#include "rand_utils.h"
#include "ICell.h"
#include "Cell_data.h"
#include <algorithm>
#include <iterator>
using namespace std;

scheduler::scheduler(const cd_list_t *ac)
	: nondone_cells(*ac),
	  cell_schedule(ac->size())
{
	assert(ac); // NOTE: this won't catch above refs to ac
	assert(ac->size() > 0);
	reschedule();
}


Cell_data* scheduler::next_cell()
{
	// Check to see if previous cell is now done
	if(next_scheduled_cell_index > 0)
	{
		const unsigned int prev_scheduled_cell_index=next_scheduled_cell_index-1;
		check_and_mark_cell_done(cell_schedule[prev_scheduled_cell_index]);
	}
	if(next_scheduled_cell_index >= num_cells_scheduled)
	{
		reschedule();
	}

	if(0 == num_cells_scheduled)
		return NULL;

	// Hand back the next cell to run
	Cell_data *next_cell = cell_schedule[next_scheduled_cell_index];
	assert(!next_cell->get_cell()->done());
	++next_scheduled_cell_index;

	return next_cell;
}

bool scheduler::all_cells_done() const
{ return 0 == nondone_cells.size(); }


void scheduler::reschedule()
{
	if(0 == nondone_cells.size())
		num_cells_scheduled = 0;
	else
	{
		assert(nondone_cells.size() <= INT_MAX-1);
		const int max_rndval = 1+static_cast<int>(nondone_cells.size());
		num_cells_scheduled  = global_randf()(max_rndval);
	}

	cd_list_t::iterator cell_schedule_end = cell_schedule.begin();
	advance(cell_schedule_end, num_cells_scheduled);

	random__sample_shuffle(nondone_cells.begin(), nondone_cells.end(),
								  cell_schedule.begin(), cell_schedule_end,
								  global_randf(),
								  (nondone_cells.size()==num_cells_scheduled));

	next_scheduled_cell_index = 0;
}

// Given a cell's index into cell_schedule, checks to see if the
// cell is done and if it is removes it from nondone_cells.
// Returns true if cell was done.
bool scheduler::check_and_mark_cell_done(Cell_data *cell)
{
	if(cell->get_cell()->done())
	{
		cd_list_t::iterator nondone_it
			= find(nondone_cells.begin(), nondone_cells.end(), cell);
		assert(nondone_it != nondone_cells.end());
		nondone_cells.erase(nondone_it);
		return true;
	}
	else
	{
		return false;
	}
}
