/* $Id: shape_cell.cpp,v 1.4 2004/05/09 01:36:20 chris Exp $
 *
 * shapsim - Amorphous shape mapping computer simulator and analyzer.
 * Copyright (C) 2004 Chris Frost <frost@cs.virginia.edu>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "shape_cell.h"
#include "ICell.h"
#include "messenger.h"
#include "shape_sensor.h"
#include "shape_sensor_factory.h"
#include "Cell_data.h"
#include "utils.h"
#include "delivered_msg.h"
#include "rand_utils.h"
#include "math.h"
#include <vector>
#include <string>
using namespace std;


//
// neighbor_shape_data
//

neighbor_shape_data::neighbor_shape_data(unsigned int _id,
													  float d,
													  float a,
													  bool os)
	: id(_id), distance(d), angle(a), on_shape(os)
{}



//
// shape_cell
//

location::coord_t g_sc_max_neighbor_dist;

// Number of levels a cell can sense in direction and distance when
// receiving messages.
// Special value: -1 means no reduction in accuracy, just use the float.
int g_sc_num_strength_levels = -1;
int g_sc_num_dir_levels      = -1;


shape_cell::shape_cell(messenger *m,
							  Cell_data *cd,
							  const shape_sensor &ss,
							  unsigned int _id)
	: ICell(m, cd), have_sampled_env(false), ssens(ss), id(_id),
	  steps_since_last_msg(0), angle_offset(global_randf()(2*M_PI+2)-1)
	// TODO: should use a float rnd number generator rather than int--^
{}


// ICell interface

void shape_cell::step()
{
	if(!have_sampled_env)
	{
		const bool display_on_shape_or_not = false;
		if(display_on_shape_or_not)
		{
			string text;
			if(ssens.on_shape())
				text += "Y";
		else
			text += "N";
			output(text);
		}

		todeliver_msg m;
		auto_ptr<descrip_msg> dm(new descrip_msg(id, ssens.on_shape()));
		m.set_msg_content(static_cast< auto_ptr<msg_content> >(dm));
		send_msg(m);

		have_sampled_env = true;
	}

	vector<delivered_msg> new_msgs = get_recved_msgs();
	if(new_msgs.size() > 0)
	{
		steps_since_last_msg = 0;
		levelize_msg_metadata(new_msgs);
		extract_and_process_descrip_msgs(new_msgs);
		assert(new_msgs.size() == 0); // we should have processed all msgs
	}
	else
	{
		++steps_since_last_msg;
	}
}

bool shape_cell::done() const
{
	return steps_since_last_msg >= post_last_msg_steps_till_done;
}


// methods for inspecting

vector<neighbor_shape_data> shape_cell::get_shape_data() const
{ return shape_data; }

bool shape_cell::on_shape() const
{ return ssens.on_shape(); }

unsigned int shape_cell::get_id() const
{ return id; }

float shape_cell::get_angle_offset() const
{ return angle_offset; }


// private member functions

void shape_cell::output(const string &text)
{
	cout << "[" << id << "]: " << text << endl;
}

void shape_cell::levelize_msg_metadata(vector<delivered_msg> &dms)
{
	const int num_dir_levels      = g_sc_num_dir_levels;
	const int num_strength_levels = g_sc_num_strength_levels;
	
	for(vector<delivered_msg>::iterator it=dms.begin(); it!=dms.end(); ++it)
	{
		delivered_msg &cur_msg = *it;

		if(num_dir_levels != -1)
		{
			const float levelized_dir
				= round_float(num_dir_levels*cur_msg.get_direction())
				  / num_dir_levels;
			cur_msg.set_direction(levelized_dir);
		}
		if(num_strength_levels != -1)
		{
			const float levelized_strength
				= round_float(num_strength_levels*cur_msg.get_strength())
				  / num_strength_levels;
			cur_msg.set_strength(levelized_strength);
		}
	}
}

#include <iostream>
void shape_cell::extract_and_process_descrip_msgs(vector<delivered_msg> &dms)
{
	// Something to check into if trying to speed up messaging:
	// This code uses the vector dms and erases msgs after it is done with them.
	// Would it be faster to use a list for msgs than a vector? The vectors
	// should be very small, how does this play with perf?

	vector<delivered_msg>::iterator it=dms.begin();
	while(it != dms.end())
	{
		delivered_msg &cur_msg = *it;

		auto_ptr<msg_content> mc(cur_msg.get_msg_content_copy());
		auto_ptr<descrip_msg> dm(dynamic_cast<descrip_msg*>(mc.release()));
		if(!dm.get())
		{
			++it;
			continue;
		}

		if(fmod(cur_msg.get_direction(), 2*M_PI) != cur_msg.get_direction())
			cerr << "WARNING: " << cur_msg.get_direction() << endl;

		neighbor_shape_data nsd(dm->id,
										g_sc_max_neighbor_dist*cur_msg.get_strength(),
										fmod(angle_offset+cur_msg.get_direction(),
											  2*M_PI),
										dm->on_shape);
		shape_data.push_back(nsd);

		it = dms.erase(it);
	}
}



//
// shape_cell_factory
//

shape_cell_factory::shape_cell_factory(const shape_sensor_factory &nssf,
													location::coord_t max_neighbor_dist,
													int num_strength_levels,
													int num_dir_levels)
	: ssf(nssf),
	  next_cell_id(0)
{
	g_sc_max_neighbor_dist   = max_neighbor_dist;
	g_sc_num_strength_levels = num_strength_levels;
	g_sc_num_dir_levels      = num_dir_levels;
}

ICell* shape_cell_factory::operator()(Cell_data *cd, messenger *msger)
{
	assert(cd && msger);
	return new shape_cell(msger, cd, ssf(cd->get_location()), next_cell_id++);
}



//
// descrip_msg
//

descrip_msg::descrip_msg()
{}

descrip_msg::descrip_msg(unsigned int _id, bool os) : id(_id), on_shape(os)
{}

descrip_msg::~descrip_msg()
{}

descrip_msg* descrip_msg::clone() const
{ return new descrip_msg(id, on_shape); }



//
// heard_all_neighbors_msg
//

heard_all_neighbors_msg::heard_all_neighbors_msg()
{}

heard_all_neighbors_msg::~heard_all_neighbors_msg()
{}

heard_all_neighbors_msg* heard_all_neighbors_msg::clone() const
{ return new heard_all_neighbors_msg(); }
