/* $Id: shape_cell_analyzer.cpp,v 1.5 2004/05/09 01:36:20 chris Exp $
 *
 * shapsim - Amorphous shape mapping computer simulator and analyzer.
 * Copyright (C) 2004 Chris Frost <frost@cs.virginia.edu>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "shape_cell_analyzer.h"
#include "shape_cell_querier.h"
#include "shape_cell.h"
#include "progress_bar.h"
#include "utils.h"
#include "rand_utils.h"
#include "pbm_toolkit.h"
#include <vector>
#include <string>
#include <fstream>
#include <functional>
#include <hash_map>
#include <hash_set>
#include <sstream>
#include <CGAL/Cartesian.h>
#include <CGAL/ch_graham_andrew.h>
#include <CGAL/Polygon_2.h>
using namespace std;


typedef CGAL::Cartesian<float> Rep;
typedef CGAL::Point_2< Rep >  Point;

bool point_in_shape(const Point &pt, const vector<Point> &pts, Rep &traits);

// Returns the angle from a to b.
// Assumes a and b received messages from eachother.
float a_to_b_angle(unsigned int              aId,
						 unsigned int              bId,
						 const vector<shape_data> &sd_vec,
						 const vector<float>      &cells_angleoffsets);

// Returns the distance from a to b.
// Assumes a recieved a message from b.
float a_to_b_distance(unsigned int              aId,
							 unsigned int              bId,
							 const vector<shape_data> &sd_vec);

// The same as a_to_b_distance(), but does not take into account the ang offset
float a_to_b_angle_without_offset(unsigned int              aId,
											 unsigned int              bId,
											 const vector<shape_data> &sd_vec);

// Given a cell to place and its refering cell,
// returns the angle to offset all angles from the to-place cell that aligns
// it with the refering cell.
float orient_cell_with_refer(unsigned int              to_place_id,
									  unsigned int              refer_id,
									  const vector<shape_data> &sd_vec,
									  const vector<float>      &cells_angleoffsets);


cell_cov::cell_cov()
	: on_shape(false)
{}

cell_cov::cell_cov(const location &l, bool o)
	: locn(l), on_shape(o)
{}

bool cell_cov::on_shape_p() const
{ return on_shape == boundary || on_shape == on; }

location cell_cov2location::operator()(const cell_cov &cell)
{ return cell.locn; }

bool offshape_cell::operator()(const cell_cov &cell)
{ return cell.on_shape == cell_cov::off; }


// possible-TODO: make use of direction information, if no info in a certain
// direction, mark as "unknown"?
bool boundary_cell(const std::vector<shape_data> &sd_vec,
						 unsigned int                   cell_index)
{
	if( !(sd_vec[cell_index].cd_valid && sd_vec[cell_index].cd.on_shape) )
		return false;

	const cell_data &cd = sd_vec[cell_index].cd;

	if(cd.nghbrs.size() == 0)
		return true;

	bool found_not_on_shape_nghbr = false;
	for(unsigned int i=0; i<cd.nghbrs.size(); ++i)
	{
		if(!cd.nghbrs[i].on_shape)
	   {
			found_not_on_shape_nghbr = true;
			break;
		}
	}
	
	return found_not_on_shape_nghbr;
}


unsigned int decrease_cell_sample_size(float percent_remain, // [0,1]
													vector<shape_data> &sd)
{
	unsigned int num_cells_to_invalidate
		= static_cast<uint>((1-percent_remain)*sd.size());
	if(percent_remain == 0)
		num_cells_to_invalidate = sd.size();

	for(unsigned int i=0; i<num_cells_to_invalidate; ++i)
	{
		// we invalidate in order, rather than randomly, because the cells
		// were randomly placed, so there is no need.
		sd[i].cd_valid = false;		
	}
	
	return num_cells_to_invalidate;
}


bool generate_connection_graph(const vector<shape_data> &sd_vec,
										 const string &out_filename,
										 bool output_progress)
{
	ofstream connect_out(out_filename.c_str());
	if(!connect_out)
		return false;

	progress_bar progress(sd_vec.size(), output_progress);
	progress.begin();

	connect_out << "graph shape_data {" << endl;

	for(vector<shape_data>::const_iterator sd_it = sd_vec.begin();
		 sd_it!=sd_vec.end();
		 ++sd_it)
	{
		const shape_data &sd = *sd_it;
		if(sd.cd_valid && sd.cd.on_shape)
		{
			/*
			connect_out << sd.cd.id
							<< " [color=black,fontcolor=white,style=filled];" << endl;
			*/

			connect_out << sd.cd.id << " -- { ";
			for(vector<neighbor_shape_data>::const_iterator
					 nghbr_it = sd.cd.nghbrs.begin();
				 nghbr_it != sd.cd.nghbrs.end();
				 ++nghbr_it)
			{
				const unsigned int nghbr_id = (*nghbr_it).id;
				const bool         nghbr_cd_valid = sd_vec[nghbr_id].cd_valid;
				if(nghbr_cd_valid && (*nghbr_it).on_shape)
					connect_out << (*nghbr_it).id << " ";
			}
			connect_out << "};" << endl;
		}
		// else, cell wasn't on the shape or we didn't get to read it

		progress.update();
	}

	connect_out << "}" << endl;
	connect_out.close();
	progress.done();
	return true;
}



bool generate_connection_stats(const vector<shape_data> &sd_vec,
										 const string             &out_filename,
										 unsigned int    &members_index_in_largest_group,
										 bool                      output_progress)
{
	// <#cells in group, #groups of this size>
	typedef hash_map<unsigned int, unsigned int> groupSize_numGroups_map;
	groupSize_numGroups_map cell_group_sizes;

	unsigned int largest_cell_group_found=0;
	members_index_in_largest_group = UINT_MAX; // UINT_MAX: none found

	progress_bar progress(sd_vec.size(), output_progress);
	progress.begin();

	vector<bool> cells_added(sd_vec.size(), false);
	for(unsigned int i=0; i<cells_added.size(); ++i)
	{
		if(!sd_vec[i].cd_valid || cells_added[i])
			continue;

		hash_set<unsigned int> nghbrs_explored;
		hash_set<unsigned int> nghbrs_to_explore;
		nghbrs_to_explore.insert(i);

		while(!nghbrs_to_explore.empty())
		{
			const unsigned int cur_cell = *nghbrs_to_explore.begin();
			nghbrs_to_explore.erase(cur_cell);
			if(!sd_vec[cur_cell].cd_valid || cells_added[cur_cell])
				continue;

			cells_added[cur_cell] = true;
			nghbrs_explored.insert(cur_cell);
			progress.update();

			const vector<neighbor_shape_data> &nghbr_shape_data
				= sd_vec[cur_cell].cd.nghbrs;
			for(unsigned int j=0; j<nghbr_shape_data.size(); ++j)
				nghbrs_to_explore.insert(nghbr_shape_data[j].id);
		}

		cell_group_sizes[nghbrs_explored.size()]++;

		if(nghbrs_explored.size() > largest_cell_group_found)
		{
			largest_cell_group_found        = nghbrs_explored.size();
			members_index_in_largest_group  = *nghbrs_explored.begin();
		}
	}

	if(UINT_MAX == members_index_in_largest_group)
		return true; // don't output anything

	ofstream out(out_filename.c_str());
	if(!out)
		return false;

	const string delim = " ";
	out << "<# connected cells>" << delim << "<# of groups with this index>"
		 << endl;
	for(groupSize_numGroups_map::iterator it=cell_group_sizes.begin();
		 it!=cell_group_sizes.end();
		 ++it)
		out << it->first << delim << it->second << endl;

	out.close();
	progress.done();
	return true;
}




void generate_sim_shape_cov(const std::vector<shape_data> &sd_vec,
									 std::vector<cell_cov>       &cell_covs,
									 bool                         output_progress)
{
	progress_bar progress(sd_vec.size(), output_progress);
	progress.begin();
	
	// set points where cells were placed
	for(vector<shape_data>::const_iterator sd_it = sd_vec.begin();
		 sd_it!=sd_vec.end();
		 ++sd_it)
	{
		const shape_data &sd = *sd_it;
		cell_cov cc;
		cc.locn = sd.sd.locn;

		// WARNING: here we make use of the knowledge that cells are
		// only invalidated in checking. Also, we assume cells report
		// the correct data. Instead, we should open the shape file
		// and compute whether the cell is on the shape or not.
		// But, this works for current uses.
		if(sd.cd.on_shape)
			cc.on_shape = cell_cov::on;
		else
			cc.on_shape = cell_cov::off;

		cell_covs.push_back(cc);
		progress.update();
	}
	progress.done();
}

float a_to_b_angle(unsigned int              aId,
						 unsigned int              bId,
						 const vector<shape_data> &sd_vec,
						 const vector<float>      &cells_angleoffsets)
{
	const float measured_a_b_angle
		= a_to_b_angle_without_offset(aId, bId, sd_vec);
	const float a_offset  = cells_angleoffsets[aId];
	const float actual_a_b_angle = fmod(measured_a_b_angle+a_offset, 2*M_PI);

	return actual_a_b_angle;
}

float a_to_b_angle_without_offset(unsigned int              aId,
											 unsigned int              bId,
											 const vector<shape_data> &sd_vec)
{
	const vector<neighbor_shape_data> &a_nghbrs = sd_vec[aId].cd.nghbrs;
	unsigned int nghbrb_index_in_a = UINT_MAX;
	for(unsigned int nghbri=0; nghbri<a_nghbrs.size(); ++nghbri)
	{
		if(bId == a_nghbrs[nghbri].id)
		{
			nghbrb_index_in_a = nghbri;
			break;
		}
	}

	assert(UINT_MAX != nghbrb_index_in_a);

	const float measured_a_b_angle = a_nghbrs[nghbrb_index_in_a].angle;

	return measured_a_b_angle;
}
											 
float a_to_b_distance(unsigned int              aId,
							 unsigned int              bId,
							 const vector<shape_data> &sd_vec)
{
	const vector<neighbor_shape_data> &a_nghbrs = sd_vec[aId].cd.nghbrs;
	unsigned int nghbrb_index_in_a = UINT_MAX;
	for(unsigned int nghbri=0; nghbri<a_nghbrs.size(); ++nghbri)
	{
		if(bId == a_nghbrs[nghbri].id)
		{
			nghbrb_index_in_a = nghbri;
			break;
		}
	}

	assert(UINT_MAX != nghbrb_index_in_a);

	const float a_b_distance = a_nghbrs[nghbrb_index_in_a].distance;

	return a_b_distance;
}

float orient_cell_with_refer(unsigned int              to_place_id,
									  unsigned int              refer_id,
									  const vector<shape_data> &sd_vec,
									  const vector<float>      &cells_angleoffsets)
{
	//return -sd_vec[to_place_id].sd.angle_offset;

	const float r_tp_angle
		= a_to_b_angle(refer_id, to_place_id, sd_vec, cells_angleoffsets);
	const float tp_r_measured_angle
		= a_to_b_angle_without_offset(to_place_id, refer_id, sd_vec);

	const float tp_r_actual_angle = r_tp_angle - M_PI;
	//	const float tp_r_actual_angle = 2*M_PI - r_tp_angle;
	const float tp_r_angle_offset = tp_r_actual_angle - tp_r_measured_angle;
	const float tp_angle_offset   = tp_r_angle_offset;

	return tp_angle_offset;
}

void place_cell(unsigned int              refed_id,
					 unsigned int              to_place_id,
					 unsigned int              to_place_nghbr_id,
					 const vector<shape_data> &sd_vec,
					 vector<bool>             &cells_placed,
					 vector<rel_location>     &cells_locns,
					 vector<float>            &cells_angleoffsets,
					 unsigned int             &num_cells_placed,
					 progress_bar             &progress)
{
	if(cells_placed[to_place_id] || !sd_vec[to_place_id].cd_valid)
		return;

	const float to_place_angle_offset
		= orient_cell_with_refer(to_place_id,refed_id,sd_vec,cells_angleoffsets);
	cells_angleoffsets[to_place_id] = to_place_angle_offset;

	const float dist  = a_to_b_distance(refed_id, to_place_id, sd_vec);
	const float angle = a_to_b_angle(refed_id, to_place_id,
												sd_vec, cells_angleoffsets);

	const rel_location rel_locn = dist_angle2rellocn(dist, angle);

	const rel_location::coord_t nghbr_x
		= cells_locns[refed_id].get_x() + rel_locn.get_x();
	const rel_location::coord_t nghbr_y
		= cells_locns[refed_id].get_y() + rel_locn.get_y();
		
	const rel_location to_place_locn(nghbr_x, nghbr_y);
	cells_locns[to_place_id] = to_place_locn;
	
	cells_placed[to_place_id] = true;
	++num_cells_placed;
	progress.update();
}
unsigned int place_cells(unsigned int             cell_id,
								 const vector<shape_data> &sd_vec,
								 vector<bool>             &cells_placed,
								 vector<rel_location>     &cells_locns,
								 progress_bar             &progress)
{
	if(!sd_vec[cell_id].cd_valid)
		return 0;

	unsigned int num_cells_placed = 0;
	
	vector<float> cells_angleoffsets(cells_locns.size(), 0);
	// pair<ref_cell_id,pair<ref_cell_nghbr_index,to_place's_id> >
	vector< pair<uint, pair<uint,uint> > > cells_to_place;

	// place initial cell
	cells_placed[cell_id] = true;
	const rel_location locn_zero(0,0);
	cells_locns[cell_id] = locn_zero;
	// Use actual offset for first cell to orient image for easier reading
	cells_angleoffsets[cell_id] = -sd_vec[cell_id].sd.angle_offset;
	++num_cells_placed;

	// add initial cells to cells_to_place
	for(unsigned int nghbr=0; nghbr<sd_vec[cell_id].cd.nghbrs.size(); ++nghbr)
	{
		const uint to_place_id = sd_vec[cell_id].cd.nghbrs[nghbr].id;
		const pair<uint,uint>              to_add_inner(nghbr, to_place_id);
		const pair<uint, pair<uint,uint> > to_add(cell_id, to_add_inner);
		cells_to_place.push_back(to_add);
	}

	while(cells_to_place.size() > 0)
	{
		const pair<uint, pair<uint,uint> > cur(cells_to_place.front());
		cells_to_place.erase(cells_to_place.begin());

		const uint ref_id      = cur.first;
		const uint ref_nghbr   = cur.second.first;
		const uint to_place_id = cur.second.second;

		if(cells_placed[to_place_id] || !sd_vec[to_place_id].cd_valid)
			continue;

		place_cell(ref_id, to_place_id, ref_nghbr,
					  sd_vec,
					  cells_placed,
					  cells_locns, cells_angleoffsets,
					  num_cells_placed,
					  progress);

		for(unsigned int nghbr=0;
			 nghbr<sd_vec[to_place_id].cd.nghbrs.size();
			 ++nghbr)
		{
			const uint nghbr_to_place_id=sd_vec[to_place_id].cd.nghbrs[nghbr].id;
			if(!cells_placed[nghbr_to_place_id]
				&& sd_vec[nghbr_to_place_id].cd_valid)
			{
				const pair<uint,uint>       to_add_inner(nghbr, nghbr_to_place_id);
				const pair<uint,pair<uint,uint> > to_add(to_place_id,to_add_inner);
				cells_to_place.push_back(to_add);
			}
		}
	}

	return num_cells_placed;
}


void generate_cell_shape_cov(const vector<shape_data> &sd_vec,
									  std::vector<cell_cov>    &cell_covs,
									  location                 &space_size,
									  bool                      output_progress)
{
	unsigned int initial_cell = 0;
	while(!sd_vec[initial_cell].cd_valid
			|| sd_vec[initial_cell].cd.nghbrs.size()==0)
		++initial_cell;

	generate_cell_shape_cov(sd_vec, initial_cell,
									cell_covs, space_size, output_progress);
}

void generate_cell_shape_cov(const std::vector<shape_data> &sd_vec,
									  unsigned int               member_of_group_to_use,
									  std::vector<cell_cov>         &cell_covs,
									  location                      &space_size,
									  bool                           output_progress)
{
	const unsigned int initial_cell = member_of_group_to_use;
	const unsigned int num_cells = sd_vec.size();
	vector<bool>         cells_placed(num_cells, false);
	vector<rel_location> cells_locns(num_cells);

	assert(initial_cell < sd_vec.size());

	progress_bar progress(num_cells, output_progress);
	progress.begin();

	const unsigned int num_cells_placed = place_cells(initial_cell,
																	  sd_vec,
																	  cells_placed,
																	  cells_locns,
																	  progress);
	progress.done();

	if(num_cells_placed != num_cells)
	{
		cout << "WARNING: there exists more than one group of cells, "
			  << "placing only cells in the given group." << endl;
	}

	// find the (x,y) offsets so that all cells have positive locns
	float min_xval=0, min_yval=0, max_xval=0, max_yval=0;
	for(unsigned int i=0; i<num_cells; ++i)
	{
		if(cells_placed[i])
		{
			const rel_location::coord_t
				x = cells_locns[i].get_x(),
				y = cells_locns[i].get_y();

			if(x < min_xval)
				min_xval = x;
			if(max_xval < x)
				max_xval = x;
			if(y < min_yval)
				min_yval = y;
			if(max_yval < y)
				max_yval = y;
		}
	}
	const float x_offset = -min_xval, y_offset = -min_yval;

	space_size.set_x(1+static_cast<unsigned int>(ceil(max_xval + x_offset)));
	space_size.set_y(1+static_cast<unsigned int>(ceil(max_yval + y_offset)));

	for(unsigned int i=0; i<num_cells; ++i)
	{
		if(cells_placed[i])
		{
			cell_cov cc;

			const unsigned int
				cell_x = static_cast<uint>(cells_locns[i].get_x() + x_offset),
				cell_y = static_cast<uint>(cells_locns[i].get_y() + y_offset);
			const location cell_locn(cell_x, cell_y);
			cc.locn = cell_locn;

			if(boundary_cell(sd_vec, i))
				cc.on_shape = cell_cov::boundary;
			else if(sd_vec[i].cd.on_shape)
				cc.on_shape = cell_cov::on;
			else
				cc.on_shape = cell_cov::off;
			
			cell_covs.push_back(cc);
		}
	}
}

bool not_boundary_cell::operator()(const cell_cov &cell)
{ return (cell_cov::boundary != cell.on_shape); }


bool output_shape_cov_coloring(const std::vector<cell_cov> &cell_cov,
										 const location              &space_size,
										 const std::string           &out_filename,
										 bool                         output_progress)
{
	const unsigned int max_val = 4;
	const ppm_pixel
		no_cell_pres(4,4,4),
		cell_off(0,3,3),
		cell_on(4,4,0),
		cell_boundary(0,0,0);

	const unsigned int image_width =static_cast<uint>(ceil(space_size.get_x()));
	const unsigned int image_height=static_cast<uint>(ceil(space_size.get_y()));
		
	vector< vector<ppm_pixel> > image;
	init_2d_vector(image, image_width, image_height, no_cell_pres);

	for(unsigned int i=0; i<cell_cov.size(); ++i)
	{
		const unsigned int
			cell_x = static_cast<uint>(round_float(cell_cov[i].locn.get_x())),
			cell_y = static_cast<uint>(round_float(cell_cov[i].locn.get_y()));

		ppm_pixel cell_pixel;
		if(cell_cov::boundary == cell_cov[i].on_shape)
			cell_pixel = cell_boundary;
		else if(cell_cov::on == cell_cov[i].on_shape)
			cell_pixel = cell_on;
		else if(cell_cov::off == cell_cov[i].on_shape)
			cell_pixel = cell_off;
		else
			assert(false);

		image[cell_x][cell_y] = cell_pixel;
	}

	return output_ppm(image, max_val, out_filename, output_progress);
}

bool calculate_shapecov_stats(const vector<cell_cov> &cell_covs,
										const location         &shapecov_size,
										const string           &shape_filename,
										const string           &out_filename)
{
	pbm pbm_image(shape_filename);
	const vector< vector<bool> >
		image(pbm_image.get_image(1+shapecov_size.get_x(),
										  1+shapecov_size.get_y()));
							
	// stats, cells_<ronshape>_<aonshape> is number of cells whose
	// reconstruction shows ronshape (in or out) when the actual was aonshape.
	unsigned int cells_in_in=0, cells_in_out=0, cells_out_out=0, cells_out_in=0;

	for(vector<cell_cov>::const_iterator it=cell_covs.begin();
		 it!=cell_covs.end();
		 ++it)
	{
		// TODO: this if is a hack to work around doing bad indexing.
		// Why does it happen?
		// Could it be the image-scaling program doesn't fit complete size?
		if(it->locn.get_y() >= image.size()
			|| it->locn.get_x() >= image[0].size())
		{
			cerr << "+";
			continue;
		}

		const bool correct
			= it->on_shape_p() == image[it->locn.get_y()][it->locn.get_x()];
		
		if(correct)
		{
			if(it->on_shape_p())
				cells_in_in++;
			else
				cells_out_out++;
		}
		else
		{
			if(it->on_shape_p())
				cells_in_out++;
			else
				cells_out_in++;
		}
	}
	const unsigned int total_num_cells
		= cells_in_in + cells_in_out + cells_out_in + cells_out_out;

	const float
		percent_in_correct   = static_cast<float>(cells_in_in)  /total_num_cells,
		percent_out_correct  = static_cast<float>(cells_out_out)/total_num_cells,
		percent_in_incorrect = static_cast<float>(cells_in_out) /total_num_cells,
		percent_out_incorrect= static_cast<float>(cells_out_in) /total_num_cells;

	stringstream ss;
	ss << "in-in: "     << percent_in_correct                           << endl
		<< "out-out: "   << percent_out_correct                          << endl
		<< "in-out: "    << percent_in_incorrect                         << endl
		<< "out-in: "    << percent_out_incorrect                        << endl
		<< "correct: "   << percent_in_correct + percent_out_correct     << endl
		<< "incorrect: " << percent_in_incorrect + percent_out_incorrect << endl;

	return output_string(ss.str(), out_filename);
}

bool output_locations(const vector<location> &locns,
							 const location         &space_size,
							 const string           &out_filename,
							 bool                    output_progress)
{
	const unsigned int max_val = 1;
	const unsigned int
		no_cell_pres(0),
		cell_pres(1);

	const unsigned int image_width
		= static_cast<unsigned int>(ceil(space_size.get_x()));
	const unsigned int image_height
		= static_cast<unsigned int>(ceil(space_size.get_y()));
	vector< vector<unsigned int> > image;
	init_2d_vector(image, image_width, image_height, no_cell_pres);

	for(unsigned int i=0; i<locns.size(); ++i)
	{
		const unsigned int
			cell_x = static_cast<uint>(locns[i].get_x()),
			cell_y = static_cast<uint>(locns[i].get_y());

		image[cell_x][cell_y] = cell_pres;
	}

	return output_pgm(image, max_val, out_filename, output_progress);
}


class location2CGALPoint : public std::unary_function<const location&, Point>
{
public:
	Point operator()(const location &l)
	{
		const Point p(l.get_x(), l.get_y());
		return p;
	}
};
class CGALPoint2location :  public std::unary_function<const Point&, location>
{
public:
	location operator()(const Point &p)
	{
		const location l(p.x(), p.y());
		return l;
	}
};

void convex_hull(const std::vector<location> &pts,
					  std::vector<location>       &hull_pts)
{

	vector<Point> Ppts;
	transform(pts.begin(), pts.end(), back_inserter(Ppts),
				 location2CGALPoint());

	vector<Point> Phull_pts;
	CGAL::ch_graham_andrew(Ppts.begin(),
								  Ppts.end(),
								  back_inserter(Phull_pts));
	transform(Phull_pts.begin(), Phull_pts.end(), back_inserter(hull_pts),
				 CGALPoint2location());
}

bool point_in_shape(const Point &pt, const vector<Point> &pts, Rep &traits)
{
	switch(CGAL::bounded_side_2(pts.begin(), pts.end(), pt, traits))
	{
		case CGAL::ON_BOUNDED_SIDE:
		case CGAL::ON_BOUNDARY:
			return true;
		case CGAL::ON_UNBOUNDED_SIDE:
			return false;
		default:
			assert(false);
	}
}

bool point_in_shape(const location &pt, const vector<location> &pts)
{
	Point Ppt = location2CGALPoint()(pt);

	vector<Point> Ppts;
	transform(pts.begin(), pts.end(), back_inserter(Ppts),location2CGALPoint());
	Rep traits = Rep();
	return point_in_shape(Ppt, Ppts, traits);
}

void calc_percent_inout(const vector<location> &sim_pts,
								const vector<location> &cell_pts,
								const location         &space_size,
								unsigned int            num_samples,
								float &correct_in,   float &correct_out,
								float &correct,
								float &incorrect_in, float &incorrect_out,
								float &incorrect,
								bool                    output_progress)
{
	
	correct_in   = 0; correct_out   = 0;
	incorrect_in = 0; incorrect_out = 0;
	unsigned int
		n_correct_in   = 0, n_correct_out   = 0,
		n_incorrect_in = 0, n_incorrect_out = 0,
		n_in           = 0, n_out           = 0;

	vector<Point> Psim_pts, Pcell_pts;
	transform(sim_pts.begin(), sim_pts.end(), back_inserter(Psim_pts),
				 location2CGALPoint());
	transform(cell_pts.begin(), cell_pts.end(), back_inserter(Pcell_pts),
				 location2CGALPoint());

	Rep traits = Rep();
	const int
		space_max_x = 1 + static_cast<uint>(space_size.get_x()),
		space_max_y = 1 + static_cast<uint>(space_size.get_y());
	assert(space_max_x >= 0 && space_max_y >= 0);

	progress_bar progress(num_samples, output_progress);
	progress.begin();

	for(unsigned int i=0; i<num_samples; ++i)
	{
		Point test_pt(global_randf()(space_max_x),
						  global_randf()(space_max_y));
		const bool
			sim_in  = point_in_shape(test_pt, Psim_pts,  traits),
			cell_in = point_in_shape(test_pt, Pcell_pts, traits);
		if(sim_in == cell_in)
		{
			if(cell_in) ++n_correct_in;
			else        ++n_correct_out;
		}
		else
		{
			if(cell_in) ++n_incorrect_in;
			else        ++n_incorrect_out;
		}
		if(sim_in) ++n_in;
		else       ++n_out;

		progress.update();
	}

	correct_in    = static_cast<float>(n_correct_in)      / n_in;
	correct_out   = static_cast<float>(n_correct_out)     / n_out;
	correct       =
		static_cast<float>(n_correct_in+n_correct_out)     / num_samples;
		
	incorrect_in  = static_cast<float>(n_incorrect_in)    / n_in;
	incorrect_out = static_cast<float>(n_incorrect_out)   / n_out;
	incorrect     =
		static_cast<float>(n_incorrect_in+n_incorrect_out) / num_samples;

	progress.done();
}

typedef CGAL::Polygon_2<Rep> Polygon;
float area(const vector<location> &pts)
{
	vector<Point> Ppts;
	transform(pts.begin(), pts.end(), back_inserter(Ppts),location2CGALPoint());
	Polygon pgn(Ppts.begin(), Ppts.end());
	return pgn.area();
}

bool output_string(const string &str, const string &filename)
{
	ofstream out(filename.c_str());
	if(!out)
		return false;

	out << str << endl;

	out.close();
	return true;
}



//
//
// pnm generation

ppm_pixel::ppm_pixel() : red(0), green(0), blue(0) {}
ppm_pixel::ppm_pixel(unsigned int r, unsigned int g, unsigned int b)
		: red(r), green(g), blue(b)
{}

ostream& operator<<(ostream &os, const ppm_pixel &p)
{
	const string delim = " ";
	os << p.red << delim << p.green << delim << p.blue;
	return os;
}


bool output_ppm(const vector< vector<ppm_pixel> > &image,
					 unsigned int                max_val,
					 const string               &out_filename,
					 bool                        output_progress)
{
	const unsigned int image_width  = image.size();
	const unsigned int image_height = image[0].size();

	const string magic_num = "P3";
	const string delim     = "  ";

	ofstream out(out_filename.c_str());
	if(!out)
		return false;

	progress_bar progress_out(image_height, output_progress);
	progress_out.begin();
	// output file
	out << magic_num << endl;
	out << image_width << delim << image_height << endl;
	out << max_val << endl;
	for(unsigned int y=0; y<image_height; ++y)
	{
		for(unsigned int x=0; x<image_width; ++x)
		{
			out << image[x][y] << delim;
		}
		out << endl;
		
		progress_out.update();
	}

	out.close();
	progress_out.done();
	return true;
}

bool output_pgm(const vector< vector<unsigned int> > &image,
					 unsigned int                max_greyval,
					 const string               &out_filename,
					 bool                        output_progress)
{
	const unsigned int image_width  = image.size();
	const unsigned int image_height = image[0].size();

	const string magic_num = "P2";
	const string delim     = " ";

	ofstream out(out_filename.c_str());
	if(!out)
		return false;

	progress_bar progress_out(image_height, output_progress);
	progress_out.begin();
	// output file
	out << magic_num << endl;
	out << image_width << delim << image_height << endl;
	out << max_greyval << endl;
	for(unsigned int y=0; y<image_height; ++y)
	{
		for(unsigned int x=0; x<image_width; ++x)
		{
			out << image[x][y] << delim;
		}
		out << endl;
		
		progress_out.update();
	}
	out.close();
	progress_out.done();

	return true;
}
