/* $Id: shape_cell_analyzer.h,v 1.4 2004/05/09 01:36:20 chris Exp $
 *
 * shapsim - Amorphous shape mapping computer simulator and analyzer.
 * Copyright (C) 2004 Chris Frost <frost@cs.virginia.edu>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifndef __SHAPE_CELL_ANALYZER_H
#define __SHAPE_CELL_ANALYZER_H

#include "shape_cell_querier.h"
#include "location.h"
#include <vector>
#include <string>
#include <algorithm>
#include <functional>

class cell_cov
{
 public:
	cell_cov();
	cell_cov(const location &l, bool o);

	location locn;
	int on_shape;
	static const int boundary=2, on=1, off=0;

	bool on_shape_p() const;
};

class cell_cov2location : public std::unary_function<const cell_cov&, location>
{ public: location operator()(const cell_cov &cell); };

class offshape_cell : public std::unary_function<const cell_cov&, bool>
{ public: bool operator()(const cell_cov &cell); };

// true iff the given cell is calculated to be a boundary cell
bool boundary_cell(const std::vector<shape_data> &sd_vec,
						 unsigned int                   cell_index);


// randomly marks shape_data::cd as invalid for 1-percent_remain % of cells
unsigned int decrease_cell_sample_size(float percent_remain, // [0,1]
													std::vector<shape_data> &sd);

// generate connection graph for processing by neato
bool generate_connection_graph(const std::vector<shape_data> &sd,
										 const std::string             &out_filename,
										 bool                           output_progress);

// generate statistics about cell connections, the distribution of number
// of cells in each group of cells
bool generate_connection_stats(const std::vector<shape_data> &sd,
										 const std::string             &out_filename,
										 unsigned int             &largest_group_member,
										 bool                           output_progress);


// generate the reconstructed shape using the simulator's internal data
void generate_sim_shape_cov(const std::vector<shape_data> &sd_vec,
									 std::vector<cell_cov>         &cell_covs,
									 bool                           output_progress);

// generate the reconstructed shape using cells' queried ata
void generate_cell_shape_cov(const std::vector<shape_data> &sd_vec,
									  unsigned int               member_of_group_to_use,
									  std::vector<cell_cov>         &cell_covs,
									  location                      &space_size,
									  bool                           output_progress);
// this version of g_c_s_c() finds the first valid cell and uses it to
// call the previous function.
void generate_cell_shape_cov(const std::vector<shape_data> &sd_vec,
									  std::vector<cell_cov>         &cell_covs,
									  location                      &space_size,
									  bool                           output_progress);

class not_boundary_cell : public std::unary_function<const cell_cov&, bool>
{ public: bool operator()(const cell_cov &cell); };

// save a generated shape_cov as an image
bool output_shape_cov_coloring(const std::vector<cell_cov> &cell_cov,
										 const location              &space_size,
										 const std::string           &out_filename,
										 bool                         output_progress);

// calculate stats about point sample description's correctness
bool calculate_shapecov_stats(const vector<cell_cov> &cell_covs,
										const location         &shapecov_size,
										const string           &shape_filename,
										const string           &out_filename);

// from the given locations, generate an image with pixels at each location
bool output_locations(const std::vector<location> &locns,
							 const location              &space_size,
							 const std::string           &out_filename,
							 bool                         output_progress);

// find the convex hull of the given points
void convex_hull(const std::vector<location> &pts,
					  std::vector<location>       &hull_pts);

bool point_in_shape(const location &pt, const std::vector<location> &pts);
void calc_percent_inout(const std::vector<location> &sim_pts,
								const std::vector<location> &cell_pts,
								const location              &space_size,
								unsigned int                 num_samples,
								float &correct_in,   float &correct_out,
								float &correct,
								float &incorrect_in, float &incorrect_out,
								float &incorrect,
								bool                         output_progress);

// find the area of the given polygon
float area(const std::vector<location> &pts);

// output a string
bool output_string(const std::string &str, const std::string &filename);


//
// pnm generation

class ppm_pixel
{
public:
	ppm_pixel();
	ppm_pixel(unsigned int r, unsigned int g, unsigned int b);
	unsigned int red, green, blue;
};
ostream& operator<<(ostream &os, const ppm_pixel &p);

template<class T>
void init_2d_vector(vector< vector<T> > &v,
						  unsigned int width, unsigned int height,
						  const T &init_val);

bool output_ppm(const std::vector< std::vector<ppm_pixel> > &image,
					 unsigned int                                 max_val,
					 const string                                &out_filename,
					 bool                                         output_progress);

bool output_pgm(const std::vector< std::vector<unsigned int> > &image,
					 unsigned int                                    max_greyval,
					 const string                                   &out_filename,
					 bool                                          output_progress);


//
// template implementation

template<class T>
void init_2d_vector(vector< vector<T> > &v,
						  unsigned int width,
						  unsigned int height,
						  const T &init_val)
{
	vector<T> row(height);
	fill(row.begin(), row.end(), init_val);

	v.resize(width);
	fill(v.begin(), v.end(), row);
}

#endif // __SHAPE_CELL_ANALYZER_H
