/* $Id: shape_cell_querier.cpp,v 1.3 2004/05/09 01:36:20 chris Exp $
 *
 * shapsim - Amorphous shape mapping computer simulator and analyzer.
 * Copyright (C) 2004 Chris Frost <frost@cs.virginia.edu>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "shape_cell_querier.h"
#include "cd_list.h"
#include "Cell_data.h"
#include "shape_cell.h"
#include "progress_bar.h"
#include <vector>
#include <fstream>
#include <functional>
#include <algorithm>
#include <string>
using namespace std;

//
// cell_data
//

cell_data::cell_data(unsigned int                            cell_id,
							bool                                    cell_on_shape,
							const std::vector<neighbor_shape_data> &nghbr_shape_data)
	: id(cell_id), on_shape(cell_on_shape), nghbrs(nghbr_shape_data)
{}

bool cell_data_comp::operator()(const cell_data &x, const cell_data &y)
{ return x.id < y.id; }


//
// sim_data
//

sim_data::sim_data(const location  &sim__cell_locn,
						 float            sim__angle_offset,
						 const cd_list_t &sim__ngbhrs)
	: locn(sim__cell_locn),
	  angle_offset(sim__angle_offset)
{
	for(cd_list_t::const_iterator it=sim__ngbhrs.begin();
		 it!=sim__ngbhrs.end();
		 ++it)
	{
		const Cell_data *cur_nghbr_cd = *it;
		const shape_cell *cur_nghbr
			= dynamic_cast<const shape_cell*>(cur_nghbr_cd->get_cell());
		nghbrs.push_back(cur_nghbr->get_id());
	}
}

sim_data::sim_data(const location &sim__cell_locn,
						 float           sim__angle_offset,
						 const vector<unsigned int> &sim__ngbhr_ids)
	:locn(sim__cell_locn),angle_offset(sim__angle_offset),nghbrs(sim__ngbhr_ids)
{}


//
// shape_data
//

shape_data::shape_data(const cell_data &_cd, const sim_data &_sd)
	: cd(_cd), cd_valid(true), sd(_sd)
{}

bool shape_data_comp::operator()(const shape_data &x, const shape_data &y)
	// TODO: handle case when cd_invalid
{ return x.cd.id < y.cd.id; }



//
// query/save/load functions
//

#include <iostream>
#include <algorithm>
void query_cells(const cd_list_t::const_iterator cells_begin,
					  const cd_list_t::const_iterator cells_end,
					  vector<shape_data> &shaped_vec,
					  bool output_progress)
{
	progress_bar progress(distance(cells_begin, cells_end), output_progress);
	progress.begin();

	for(cd_list_t::const_iterator cell_it=cells_begin;
		 cell_it!=cells_end;
		 ++cell_it)
	{
		const Cell_data  *cd   = *cell_it;
		const shape_cell *cell = dynamic_cast<const shape_cell*>(cd->get_cell());
		assert(cell);
	
		cell_data  celld(cell->get_id(),
							  cell->on_shape(),
							  cell->get_shape_data());
		sim_data   simd(cd->get_location(),
							 cell->get_angle_offset(),
							 *(cd->get_neighbors()));
		shape_data shaped(celld, simd);

		shaped_vec.push_back(shaped);

		progress.update();
	}

	progress.done();

	shape_data_comp comp;
	sort(shaped_vec.begin(), shaped_vec.end(), comp);
}


// TODO: add locn_nghbr_filename and shape_descrip_filename to file format

// file format:
// <num_cells>
// <space_width> <space_height>
// <neighbor_max_dist>
// <prng_seed>
//
//
// <cell_data_valid>
// <id> <on_shape>
// <num_nghbrs>
// <nghbr_1 id> <nghbr_1 dist> <nghbr_1 angle> <nghbr_1 on_shape>
// <nghbr_2 id> ...
// <sim_location_x> <sim_location_y>
// <sim_cell_angle_offset>
// <sim_num_nghbrs>
// <sim_nghbr_1 id> <sim_nghbr_2 id> ...
//
// <cell_data_valid>
// ...

const string delim = " ";

bool save_queried_data(const string             &filename,
							  const vector<shape_data> &shaped_vec,
							  const location           &space_size,
							  const location::coord_t   neighbor_max_dist,
							  const string             &locn_nghbr_filename,
							  const string             &shape_descrip_filename,
							  unsigned int              prng_seed,
							  bool                      output_progress)
{
	const unsigned int num_cells = shaped_vec.size();
	ofstream out(filename.c_str());
	if(!out)
		return false;

	out << num_cells << endl
		 << space_size.get_x() << delim << space_size.get_y() << endl
		 << neighbor_max_dist << endl
		 << prng_seed << endl
		 << endl;
		
	progress_bar progress(num_cells, output_progress);
	progress.begin();

	for(unsigned int i=0; i<num_cells; ++i)
	{
		out << shaped_vec[i].cd_valid << endl;
		out << shaped_vec[i].cd.id << delim << shaped_vec[i].cd.on_shape << endl;
		out << shaped_vec[i].cd.nghbrs.size() << endl;
		for(unsigned int j=0; j<shaped_vec[i].cd.nghbrs.size(); ++j)
		{
			out << shaped_vec[i].cd.nghbrs[j].id               << delim
				 << shaped_vec[i].cd.nghbrs[j].distance         << delim
				 << shaped_vec[i].cd.nghbrs[j].angle            << delim
				 << shaped_vec[i].cd.nghbrs[j].on_shape         << endl;				 
		}

		out << shaped_vec[i].sd.locn.get_x() << delim
			 << shaped_vec[i].sd.locn.get_y() << endl;
		out << shaped_vec[i].sd.angle_offset << endl;
		out << shaped_vec[i].sd.nghbrs.size() << endl;
		for(unsigned int j=0; j<shaped_vec[i].sd.nghbrs.size(); ++j)
		{
			out << shaped_vec[i].sd.nghbrs[j] << delim;
		}
		out << endl << endl;

		progress.update();
	}

	out.close();
	progress.done();
	return true;
}

bool load_queried_data(const string       &filename,
							  vector<shape_data> &shaped_vec,
							  location           &space_size,
							  location::coord_t  &neighbor_max_dist,
							  string             &locn_nghbr_filename,
							  string             &shape_descrip_filename,
							  unsigned int        prng_seed,
							  bool                output_progress)
{
	ifstream in(filename.c_str());
	if(!in)
		return false;

	unsigned int num_cells;
	location::coord_t space_width, space_height;
	in >> num_cells
		>> space_width >> space_height
		>> neighbor_max_dist
		>> prng_seed;
	space_size.set_x(space_width);
	space_size.set_y(space_height);

	progress_bar progress(num_cells, output_progress);
	progress.begin();

	for(unsigned int i=0; i<num_cells; ++i)
	{
		// load cell_data
		bool cd_valid;
		unsigned int id;
		bool on_shape;
		in >> cd_valid >> id >> on_shape;

		unsigned int num_nghbrs;
		in >> num_nghbrs;
		vector<neighbor_shape_data> nsd_vec;
		for(unsigned int j=0; j<num_nghbrs; ++j)
		{
			unsigned int nghbr_id;
			float nghbr_dist, nghbr_angle;
			bool nghbr_on_shape;
			in >> nghbr_id >> nghbr_dist >> nghbr_angle >> nghbr_on_shape;
			
			const neighbor_shape_data nsd(nghbr_id,
													nghbr_dist,
													nghbr_angle,
													nghbr_on_shape);
			nsd_vec.push_back(nsd);
		}
		const cell_data celld(id, on_shape, nsd_vec);

		// load sim_data
		location::coord_t sim_x, sim_y;
		in >> sim_x >> sim_y;
		const location sim_locn(sim_x, sim_y);

		float sim_angle_offset;
		in >> sim_angle_offset;

		unsigned int sim_num_nghbrs;
		in >> sim_num_nghbrs;
		vector<unsigned int> sim_nghbrs;
		for(unsigned int j=0; j<sim_num_nghbrs; ++j)
		{
			unsigned int sim_nghbr_id;
			in >> sim_nghbr_id;
			sim_nghbrs.push_back(sim_nghbr_id);
		}
		const sim_data simd(sim_locn, sim_angle_offset, sim_nghbrs);

		// store loaded data
		const shape_data shaped(celld, simd);
		shaped_vec.push_back(shaped);

		progress.update();
	}

	in.close();
	progress.done();
	return true;
}
