/* $Id: simulator_impl.h,v 1.5 2004/05/11 12:42:49 chris Exp $
 *
 * shapsim - Amorphous shape mapping computer simulator and analyzer.
 * Copyright (C) 2004 Chris Frost <frost@cs.virginia.edu>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifndef __SIMULATOR_IMPL_H
#define __SIMULATOR_IMPL_H

//
// Implementation of class simulator
//

#include "simulator.h"
#include "todeliver_msg.h"
#include "delivered_msg.h"
#include "msg_content.h"
#include "location.h"
#include "cd_list.h"
#include "rand_utils.h"
#include "location_neighbor_generator.h"
#include <algorithm>
using namespace std;

template<class CellFactory>
simulator<CellFactory>::simulator(const string &locn_nghbr_filename,
											 const location &_space_size,
											 const location::coord_t nmd,
											 CellFactory cf)
	: cell_fact(cf),
	  cell_scheduler(NULL),
	  neighbor_max_dist(nmd)
{
	assert(nmd >= 0);
	const bool verbose_output = true;
	const string init_indent = string("   ") + "   ";
	
	const location_neighbor_generator lng;

	location space_size;
	location::coord_t loader_nmd;
	
	// Load location and nghbr data from file
	if(verbose_output)
		cout << init_indent << "loading location and neighbor data from file:"
			  << endl;
	location_neighbor_generator::location_list_t   locns;
	location_neighbor_generator::nghbr_list_list_t nghbrs;
	assert(lng.load(locn_nghbr_filename, space_size, loader_nmd,
						 locns, nghbrs, false));
	const unsigned int num_cells = locns.size();

	// make sure the data we loaded agrees with what the caller is using:
	assert(space_size.get_x() == _space_size.get_x()
			 && space_size.get_y() == _space_size.get_y());
	assert(neighbor_max_dist == loader_nmd);
	

	// Create and set Cell_data locns and nghbrs
	if(verbose_output)
		cout << init_indent << "creating cell_data objects" << endl;
	for(unsigned int i=0; i<num_cells; ++i)
	{
		Cell_data *cd = new Cell_data();
		cd->set_location(locns[i]);
		cells.push_back(cd);
	}
	for(unsigned int i=0; i<num_cells; ++i)
	{
		cd_list_t cell_i_nghbrs;
		for(unsigned int j=0; j<nghbrs[i].size(); ++j)
			cell_i_nghbrs.push_back(cells[nghbrs[i][j]]);
		cells[i]->set_neighbors(cell_i_nghbrs);
	}

	// Create the cells
	if(verbose_output)
		cout << init_indent << "creating cells" << endl;
	for(cd_list_t::iterator it=cells.begin(); it != cells.end(); ++it)
	{
		ICell *cell = cell_fact(*it, this);
		(*it)->set_cell(cell);
	}

	if(verbose_output)
		cout << init_indent << "initing scheduler" << endl;
	auto_ptr<scheduler> new_cell_scheduler(new scheduler(&cells));
	cell_scheduler = new_cell_scheduler;
}

template<class CellFactory>
simulator<CellFactory>::~simulator()
{
	for(cd_list_t::iterator it = cells.begin(); it != cells.end(); ++it)
	{
		delete *it;
		*it = NULL;
	}
	cells.clear();
}


template<class CellFactory>
void simulator<CellFactory>::run()
{
	Cell_data *next_cell(NULL);
	while( (next_cell = cell_scheduler->next_cell()) )
		next_cell->get_cell()->step();
}

template<class CellFactory>
const cd_list_t* simulator<CellFactory>::get_cells() const
{ return &cells; }

template<class CellFactory>
cd_list_t* simulator<CellFactory>::get_cells()
{ return &cells; }


// messenger interface
template<class CellFactory>
void simulator<CellFactory>::deliver_msg(const todeliver_msg &msg,
													  Cell_data *cd)
{
	assert(cd);

   cd_list_t      *neighbors = cd->get_neighbors();
	const location  sender_locn(cd->get_location());
	delivered_msg   template_dm;
	template_dm.set_msg_content(msg.get_msg_content_copy());

	for(cd_list_t::iterator it=neighbors->begin(); it != neighbors->end(); ++it)
	{
		Cell_data    *recipient = (*it);
		delivered_msg recipients_msg(template_dm);
		
		const float dist = locn_distance(recipient->get_location(), sender_locn);
		const float zerof=0, onef=1;
		const float strength = min(onef, max(zerof, dist / neighbor_max_dist));
		recipients_msg.set_strength(strength);
		
		const float dir = locn_direction(recipient->get_location(), sender_locn);
		recipients_msg.set_direction(dir);
		
		recipient->get_cell()->recv_msg(recipients_msg);
	}
}

#endif // __SIMULATOR_IMPL_H
